#include <testutil/internal.hpp>

namespace glmnetpp {

constexpr double InternalParams::sml;
constexpr double InternalParams::eps;
constexpr double InternalParams::big;
constexpr int InternalParams::mnlam;
constexpr double InternalParams::rsqmax;
constexpr double InternalParams::pmin;
constexpr double InternalParams::exmx;
constexpr int InternalParams::itrace;
constexpr double InternalParams::bnorm_thr;
constexpr int InternalParams::bnorm_mxit;

} // namespace glmnetpp
