#pragma once
#include <glmnetpp_bits/elnet_path/base.hpp>
#include <glmnetpp_bits/elnet_path/gaussian_base.hpp>
#include <glmnetpp_bits/elnet_path/gaussian_cov.hpp>
#include <glmnetpp_bits/elnet_path/gaussian_naive.hpp>
#include <glmnetpp_bits/elnet_path/gaussian_multi.hpp>
#include <glmnetpp_bits/elnet_path/sp_gaussian_cov.hpp>
#include <glmnetpp_bits/elnet_path/sp_gaussian_naive.hpp>
#include <glmnetpp_bits/elnet_path/sp_gaussian_multi.hpp>

#include <glmnetpp_bits/elnet_path/binomial_base.hpp>
#include <glmnetpp_bits/elnet_path/binomial_two_class.hpp>
#include <glmnetpp_bits/elnet_path/binomial_multi_class.hpp>
#include <glmnetpp_bits/elnet_path/binomial_multi_class_group.hpp>
#include <glmnetpp_bits/elnet_path/sp_binomial_two_class.hpp>
#include <glmnetpp_bits/elnet_path/sp_binomial_multi_class.hpp>
#include <glmnetpp_bits/elnet_path/sp_binomial_multi_class_group.hpp>

#include <glmnetpp_bits/elnet_path/poisson_base.hpp>
#include <glmnetpp_bits/elnet_path/poisson_naive.hpp>
#include <glmnetpp_bits/elnet_path/sp_poisson_naive.hpp>
